import atexit
import functools
import json
import os
import pathlib
import shutil
import sqlite3
import tempfile
from collections import defaultdict

import llm
import pydantic
from dotenv import dotenv_values


def get_config_var(var_name: str, default):
    return dotenv_values().get(var_name, os.environ.get(var_name, default))

class Model(pydantic.BaseModel):
    name: str
    context_length: int

class _ConfigValueNotFound():
    pass

ConfigValueNotFound = _ConfigValueNotFound()

class _ConfigHandler():
    # Default fallbacks for variables defined in either .env or environment
    _config_var_defaults = {
        "DATABASE_PATH": "fossil.db",
        "OPENAI_KEY": "",
        "OPENAI_API_BASE": "https://api.openai.com/v1",
        "MASTO_BASE": "https://hachyderm.io",
    }
    
    _model_lengths = defaultdict(
        lambda: 10000, 
        {"gpt-3.5-turbo": 4097, "ada-002": 8191}
    )
    
    _model_cache = {}
    
    def __getattr__(self, item: str):
        c_val = get_config_var(item, self._config_var_defaults.get(item, ConfigValueNotFound))
        
        if isinstance(c_val, _ConfigValueNotFound):
            raise AttributeError(f"{item} is not defined in either the enviroment or .env file")
        return c_val

    def _get_from_session(self, session_id: str| None, item: str) -> str:
        if not session_id:
            return ""
        with sqlite3.connect(self.DATABASE_PATH) as conn:
            c = conn.cursor()
            c.execute('SELECT settings FROM sessions WHERE id = ?', [session_id])
            row = c.fetchone()
            try:
                return json.loads(row[0]).get(item, "")
            except (json.decoder.JSONDecodeError, IndexError, TypeError):
                return ""
    
    def EMBEDDING_MODEL(self, session_id: str|None = None) -> Model:
        c_val = self._get_from_session(session_id, "embedding_model")
        if not c_val:
            c_val = get_config_var("EMBEDDING_MODEL", "ada-002")
        
        if c_val not in self._model_cache:
            self._model_cache[c_val] = Model(name=c_val, context_length=self._model_lengths[c_val])
        
        return self._model_cache[c_val]
    
    def SUMMARIZE_MODEL(self, session_id: str|None = None) -> Model:
        c_val = self._get_from_session(session_id, "summarize_model")
        if not c_val:
            c_val = get_config_var("SUMMARIZE_MODEL", "gpt-3.5-turbo")
            
        if c_val not in self._model_cache:
            self._model_cache[c_val] = Model(name=c_val, context_length=self._model_lengths[c_val])
        
        return self._model_cache[c_val]


ConfigHandler = _ConfigHandler()


def headers():
    return {"Authorization": f"Bearer {ConfigHandler.ACCESS_TOKEN}"}

def get_installed_llms() -> set[str]:
    return {m.model.model_id for m in llm.get_models_with_aliases()}

def get_installed_embedding_models() -> set[str]:
    return {m.model.model_id for m in llm.get_embedding_models_with_aliases()}


# Static files
class StaticFiles(pydantic.BaseModel):
    """
    This manages static files so that the user can `pip install fossil-mastodon` and it runs
    fine. 

    This copies all files into a temp directory and then deletes them as the program exits. This 
    seems to work fine even in the dev workflow, since this module gets re-run every time
    uvicorn reloads the server.
    """
    class Config:
        arbitrary_types_allowed = True
    base_path: pathlib.Path
    assets_path: pathlib.Path
    templates_path: pathlib.Path
    temp_dir: tempfile.TemporaryDirectory

    @classmethod
    def from_env(cls) -> "StaticFiles":
        src_path = pathlib.Path(__file__).parent / "app"
        temp_dir = tempfile.TemporaryDirectory()
        dst_path = pathlib.Path(temp_dir.name)
        shutil.copytree(src_path / "static", dst_path / "static")
        shutil.copytree(src_path / "templates", dst_path / "templates")
        
        obj = cls(
            base_path=dst_path,
            assets_path=dst_path / "static",
            templates_path=dst_path / "templates",
            temp_dir=temp_dir,
        )

        atexit.register(obj.cleanup)

        return obj

    def cleanup(self):
        self.temp_dir.cleanup()

    def __del__(self):
        self.cleanup()


ASSETS = StaticFiles.from_env()


import abc
import datetime
import pickle
import sqlite3
import typing

import pydantic
from fastapi import Response, responses

from fossil_mastodon import config, core
if typing.TYPE_CHECKING:
    from fossil_mastodon import plugins


class Renderable(abc.ABC):
    """
    A base class for a "shape" of data to be rendered as HTML.
    """
    @abc.abstractmethod
    def render(self, **response_args) -> Response:
        """
        Render this object as a FastAPI Response.

        :param response_args: Additional arguments to pass to the Response constructor.
        """
        raise NotImplementedError()


class TrainContext(pydantic.BaseModel):
    """
    A context object for training a model. This is passed to train().
    """
    end_time: datetime.datetime
    timedelta: datetime.timedelta
    session_id: str

    def get_toots(self) -> list[core.Toot]:
        return core.Toot.get_toots_since(self.end_time - self.timedelta)

    def sqlite_connection(self) -> sqlite3.Connection:
        return sqlite3.connect(config.ConfigHandler.DATABASE_PATH)


class BaseAlgorithm(abc.ABC):
    """
    Base class for an algorithms that render your timeline. You should implemnet
    this class to create your own algorithm.

    Abstract methods:
    - render: Run the model
    - train: Train the model

    Additionally, you may want to override this method to provide a custom UI for
    your algorithm:

    - render_model_params

    Note that objects of this class must be serializable, via pickle. However, you
    can control how serialization works by overriding these methods:

    - serialize
    - deserialize
    """

    @abc.abstractmethod
    def render(self, toots: list[core.Toot], context: "plugins.RenderContext") -> Renderable:
        """
        Run the model and return a Renderable object. This object is typically
        deserialized before this method is called.

        :param toots: The toots to run the model on. This is typically 1 day of toots,
            or 6 hours, or whatever the user (you) has selected.

        :param context: A RenderContext object that you can use to render HTML. This
            is generally just passed to the Renderable object you return.
        """
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def train(cls, context: TrainContext, http_args: dict[str, str]) -> "BaseAlgorithm":
        """
        Create an instance of this algorithm, and train it on the given toots.

        :param context: Context object where training data can be obtained.
        """
        raise NotImplementedError()

    @classmethod
    def render_model_params(cls, context: "plugins.RenderContext") -> Response:
        """
        Optionally, you can render HTML input elements that capture http_args passed 
        to train(). This is useful if your agorithm has hyperparameters that you want
        to experiment with.
        """
        return responses.HTMLResponse("")

    def serialize(self) -> bytes:
        return pickle.dumps(self)
    
    @staticmethod
    def deserialize(data: bytes) -> "BaseAlgorithm":
        return pickle.loads(data)


import datetime
import functools
import importlib
import json
import logging
import sqlite3
import traceback
import typing
from typing import Optional, Type

import html2text
import llm
import numpy as np
import requests

from fossil_mastodon import config

if typing.TYPE_CHECKING:
    from fossil_mastodon.algorithm import base

import os

from pydantic import BaseModel

logger = logging.getLogger(__name__)


@functools.cache
def create_database():
    if os.path.exists(config.ConfigHandler.DATABASE_PATH):
        return

    print("Creating database")
    with sqlite3.connect(config.ConfigHandler.DATABASE_PATH) as conn:
        c = conn.cursor()

        # Create the toots table if it doesn't exist
        c.execute('''
            CREATE TABLE IF NOT EXISTS toots (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                content TEXT,
                author TEXT,
                url TEXT,
                created_at DATETIME,
                embedding BLOB,
                orig_json TEXT,
                cluster TEXT  -- Added cluster column
            )
        ''')

        conn.commit()


@functools.lru_cache()
def _get_json(toot: "Toot") -> dict:
    # meh, this isn't great, but it works
    import json
    return json.loads(toot.orig_json)


class MediaAttatchment(BaseModel):
    type: str | None
    preview_url: str | None
    url: str | None


class Toot(BaseModel):
    class Config:
        arbitrary_types_allowed = True
    id: int | None = None
    content: str | None
    author: str | None
    url: str | None
    created_at: datetime.datetime
    embedding: np.ndarray | None = None
    orig_json: str | None = None
    cluster: str | None = None  # Added cluster property

    @property
    def orig_dict(self) -> dict:
        return _get_json(self)

    @property
    def avatar_url(self) -> str | None:
        return self.orig_dict.get("account", {}).get("avatar")
    
    @property
    def profile_url(self) -> str | None:    
        return self.orig_dict.get("account", {}).get("url")

    @property
    def display_name(self) -> str | None:
        return self.orig_dict.get("account", {}).get("display_name")

    @property
    def toot_id(self) -> str | None:
        return self.orig_dict.get("id")

    @property
    def is_reply(self) -> bool:
        return self.orig_dict.get("in_reply_to_id") is not None

    @property
    def media_attachments(self) -> list[MediaAttatchment]:
        return [MediaAttatchment(type=m.get("type"), url=m.get("url"), preview_url=m.get("preview_url")) 
                for m in self.orig_dict.get("media_attachments", [])]

    @property
    def card_preview_url(self) -> str | None:
        return self.orig_dict.get("card", {}).get("image")

    @property
    def card_url(self) -> str | None:
        return self.orig_dict.get("card", {}).get("url")

    def __hash__(self):
        return hash(self.url)

    def __eq__(self, other):
        return self.url == other.url

    def save(self, init_conn: sqlite3.Connection | None = None) -> bool:
        try:
            if init_conn is None:
                conn = sqlite3.connect(config.ConfigHandler.DATABASE_PATH)
            else:
                conn = init_conn
            create_database()
            c = conn.cursor()

            # Check if the URL already exists
            c.execute('''
                SELECT COUNT(*) FROM toots WHERE url = ? and embedding is not null
            ''', (self.url,))

            result = c.fetchone()
            url_exists = result[0] > 0

            if url_exists:
                # URL already exists, handle accordingly
                return False

            c.execute('''
                DELETE FROM toots WHERE url = ?
            ''', (self.url,))

            embedding = self.embedding.tobytes() if self.embedding is not None else bytes()
            c.execute('''
                INSERT INTO toots (content, author, url, created_at, embedding, orig_json, cluster)
                VALUES (?, ?, ?, ?, ?, ?, ?)
            ''', (self.content, self.author, self.url, self.created_at, embedding, self.orig_json, self.cluster))

        except:
            conn.rollback()
            raise
        finally:
            if init_conn is None:
                conn.commit()
        return True

    @classmethod
    def get_toots_since(cls, since: datetime.datetime) -> list["Toot"]:
        create_database()
        with sqlite3.connect(config.ConfigHandler.DATABASE_PATH) as conn:
            c = conn.cursor()

            c.execute('''
                SELECT 
                    id, content, author, url, created_at, embedding, orig_json, cluster
                FROM toots WHERE created_at >= ?
            ''', (since,))

            rows = c.fetchall()
            toots = []
            for row in rows:
                toot = cls(
                    id=row[0],
                    content=row[1],
                    author=row[2],
                    url=row[3],
                    created_at=row[4],
                    embedding=np.frombuffer(row[5]) if row[5] else None,
                    orig_json=row[6],
                    cluster=row[7]  # Added cluster property
                )
                toots.append(toot)

            return toots

    @classmethod
    def get_by_id(cls, id: int) -> Optional["Toot"]:
        create_database()
        with sqlite3.connect(config.ConfigHandler.DATABASE_PATH) as conn:
            c = conn.cursor()

            c.execute('''
                SELECT 
                    id, content, author, url, created_at, embedding, orig_json, cluster
                FROM toots WHERE id = ?
            ''', (id,))

            row = c.fetchone()
            if row:
                toot = cls(
                    id=row[0],
                    content=row[1],
                    author=row[2],
                    url=row[3],
                    created_at=row[4],
                    embedding=np.frombuffer(row[5]) if row[5] else None,
                    orig_json=row[6],
                    cluster=row[7],  # Added cluster property
                )
                return toot
            return None

    @staticmethod
    def get_latest_date() -> datetime.datetime | None:
        create_database()
        with sqlite3.connect(config.ConfigHandler.DATABASE_PATH) as conn:
            c = conn.cursor()

            c.execute('''
                SELECT MAX(created_at) FROM toots
            ''')

            result = c.fetchone()
            latest_date = result[0] if result[0] else None

            if isinstance(latest_date, str):
                try:
                    latest_date = datetime.datetime.strptime(latest_date, "%Y-%m-%d %H:%M:%S.%f")
                except ValueError:
                    latest_date = datetime.datetime.strptime(latest_date, "%Y-%m-%d %H:%M:%S")
            return latest_date

    @classmethod
    def from_dict(cls, data):
        import json

        if data.get("reblog"):
            return cls.from_dict(data["reblog"])

        return cls(
            content=data.get("content"),
            author=data.get("account", {}).get("acct"),
            url=data.get("url"),
            created_at=datetime.datetime.strptime(data.get("created_at"), "%Y-%m-%dT%H:%M:%S.%fZ"),
            orig_json=json.dumps(data),
        )

    def do_star(self):
        print("star", self.url)

    def do_boost(self):
        print("boost", self.url)


def get_toots_since(since: datetime.datetime, session_id: str):
    assert isinstance(since, datetime.datetime), type(since)
    create_database()
    download_timeline(since, session_id)
    return Toot.get_toots_since(since)


def download_timeline(since: datetime.datetime, session_id: str):
    last_date = Toot.get_latest_date()
    logger.info(f"last toot date: {last_date}")
    last_date = last_date or since
    earliest_date = None
    buffer: list[Toot] = []
    last_id = ""
    curr_url = f"{config.ConfigHandler.MASTO_BASE}/api/v1/timelines/home?limit=40"
    import json as JSON
    while not earliest_date or earliest_date > last_date:
        response = requests.get(curr_url, headers=config.headers())
        response.raise_for_status()
        json = response.json()
        if not json:
            logger.info("No more toots")
            break
        if len(json) > 1:
            last_id = json[-1]["id"]
        logger.info(f"Got {len(json)} toots; earliest={earliest_date.isoformat() if earliest_date else None}, last_id={last_id}")
        for toot_dict in json:
            toot = Toot.from_dict(toot_dict)
            earliest_date = toot.created_at if not earliest_date else min(earliest_date, datetime.datetime.strptime(toot_dict["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"))
            buffer.append(toot)

        if "next" in response.links:
            curr_url = response.links["next"]["url"]
        else:
            break
    logger.info(f"done with toots; earliest={earliest_date.isoformat() if earliest_date else None}, last_date: {last_date.isoformat() if last_date else None}")

    page_size = 50
    if len(buffer) > 0:
        num_pages = len(buffer) // page_size + 1
    else:
        num_pages = 0
    for page in range(num_pages):
        start_index = page * page_size
        end_index = start_index + page_size
        page_toots = buffer[start_index:end_index]

        # Example: Call the _create_embeddings function
        _create_embeddings(page_toots, session_id)
        with sqlite3.connect(config.ConfigHandler.DATABASE_PATH) as conn:
            for toot in page_toots:
                toot.save(init_conn=conn)


def _create_embeddings(toots: list[Toot], session_id: str):
    # Convert the list of toots to a single string
    toots = [t for t in toots if t.content]

    # Call the llm embedding API to create embeddings
    emb_model = llm.get_embedding_model(config.ConfigHandler.EMBEDDING_MODEL(session_id).name)
    embeddings = list(emb_model.embed_batch([html2text.html2text(t.content) for t in toots]))

    # Extract the embeddings from the API response
    print(f"got {len(embeddings)} embeddings")
    for i, toot in enumerate(toots):
        toot.embedding = np.array(embeddings[i])

    # Return the embeddings
    return toots


@functools.lru_cache()
def _create_session_table():
    create_database()
    with sqlite3.connect(config.ConfigHandler.DATABASE_PATH) as conn:
        c = conn.cursor()

        # Create the toots table if it doesn't exist
        c.execute('''
            CREATE TABLE IF NOT EXISTS sessions (
                id TEXT PRIMARY KEY,
                algorithm_spec TEXT,
                algorithm BLOB,
                ui_settings TEXT
            )
        ''')

        try:
            c.execute('''
                ALTER TABLE sessions ADD COLUMN settings TEXT
            ''')
        except sqlite3.OperationalError:
            pass

        conn.commit()


class Settings(BaseModel):
    embedding_model: str | None = None
    summarize_model: str | None = None


class Session(BaseModel):
    id: str
    algorithm_spec: str | None = None
    algorithm: bytes | None = None
    ui_settings: str | None = None
    settings: Settings

    def set_ui_settings(self, ui_settings: dict[str, str]):
        self.ui_settings = json.dumps(ui_settings)
        self.save()

    def get_ui_settings(self) -> dict[str, str]:
        return json.loads(self.ui_settings or "{}")

    def get_algorithm_type(self) -> Type["base.BaseAlgorithm"] | None:
        try:
            spec = json.loads(self.algorithm_spec) if self.algorithm_spec else {}
            if "module" in spec and "class_name" in spec:
                mod = importlib.import_module(spec["module"])
                return getattr(mod, spec["class_name"])
            return None
        except ModuleNotFoundError:
            traceback.print_exc()
            return None

    def set_algorithm_by_name(self, name: str) -> Type["base.BaseAlgorithm"] | None:
        from fossil_mastodon.algorithm import base
        algo = next((algo for algo in base.get_algorithms() if algo.get_name() == name), None)
        self.algorithm_spec = json.dumps({
            "module": algo.__module__,
            "class_name": algo.__name__,
            "kwargs": {},
        })
        self.algorithm = None
        self.save()
        return algo

    @classmethod
    def get_by_id(cls, id: str) -> Optional["Session"]:
        create_database()
        _create_session_table()
        with sqlite3.connect(config.ConfigHandler.DATABASE_PATH) as conn:
            c = conn.cursor()

            c.execute('''
                SELECT id, algorithm_spec, algorithm, ui_settings, settings FROM sessions WHERE id = ?
            ''', (id,))

            row = c.fetchone()
            if row:
                session = cls(
                    id=row[0],
                    algorithm_spec=row[1],
                    algorithm=row[2],
                    ui_settings=row[3],
                    settings=Settings(**json.loads(row[4] or "{}")),
                )
                return session
            return None

    @classmethod
    def create(cls) -> "Session":
        import uuid
        return cls(id=str(uuid.uuid4()).replace("-", ""), settings=Settings())

    def save(self, init_conn: sqlite3.Connection | None = None) -> bool:
        _create_session_table()
        try:
            if init_conn is None:
                conn = sqlite3.connect(config.ConfigHandler.DATABASE_PATH)
            else:
                conn = init_conn
            create_database()
            c = conn.cursor()

            c.execute('''
                INSERT INTO sessions (id, algorithm_spec, algorithm, ui_settings, settings)
                VALUES (?, ?, ?, ?, ?)
                ON CONFLICT(id) DO UPDATE 
                    SET algorithm_spec = excluded.algorithm_spec
                      , algorithm = excluded.algorithm
                      , ui_settings = excluded.ui_settings
                      , settings = excluded.settings
            ''', (self.id, self.algorithm_spec, self.algorithm, self.ui_settings, self.settings.model_dump_json()))

            if init_conn is None:
                conn.commit()
        except:
            conn.rollback()
            raise
        return True



import llm
import numpy as np
import openai
import tiktoken
from sklearn.cluster import KMeans

from . import config, core


def assign_clusters(session_id: str, toots: list[core.Toot], n_clusters: int = 5):
    # meh, ignore toots without content. I think this might be just an image, not sure
    toots = [toot for toot in toots if toot.embedding is not None]

    # Perform k-means clustering on the embeddings
    embeddings = np.array([toot.embedding for toot in toots])
    kmeans = KMeans(n_clusters=n_clusters)
    cluster_labels = kmeans.fit_predict(embeddings)

    client = openai.OpenAI(api_key=config.ConfigHandler.OPENAI_KEY)
    for i_clusters in range(n_clusters):
        clustered_toots = [toot for toot, cluster_label in zip(toots, cluster_labels) if cluster_label == i_clusters]
        combined_text = "\n\n".join([toot.content for toot in clustered_toots])

        # Use GPT-3.5-turbo to summarize the combined text
        prompt = f"Create a single label that describes all of these related tweets, make it succinct but descriptive. The label should describe all {len(clustered_toots)} of these\n\n{combined_text}"
        model = llm.get_model(config.ConfigHandler.SUMMARIZE_MODEL(session_id).name)
        summary = model.prompt(prompt).text()

        # Do something with the summary
        for toot, cluster_label in zip(toots, cluster_labels):
            if cluster_label == i_clusters:
                toot.cluster = summary

def get_encoding(session_id: str):
    try:
        return tiktoken.encoding_for_model(config.ConfigHandler.SUMMARIZE_MODEL(session_id).name)
    except KeyError:
        encoding_name = tiktoken.list_encoding_names()[-1]
        return tiktoken.get_encoding(encoding_name)

def reduce_size(session_id: str, text: str, model_limit: int = -1, est_output_size: int = 500) -> str:
    if model_limit < 0:
        config.ConfigHandler.SUMMARIZE_MODEL(session_id).context_length
    tokens = get_encoding(session_id).encode(text)
    return get_encoding(session_id).decode(tokens[:model_limit - est_output_size])



import datetime
import re
import urllib.parse

import pydantic
import streamlit as st

from . import config, core


def get_time_frame() -> datetime.timedelta:
    time_frame = st.radio("Show last:", ["6 hours", "day", "week"], horizontal=True)
    
    if time_frame == "6 hours":
        return datetime.timedelta(hours=6)
    elif time_frame == "day":
        return datetime.timedelta(days=1)
    elif time_frame == "week":
        return datetime.timedelta(weeks=1)
    raise ValueError("Invalid time frame")


def time_ago(dt: datetime.datetime) -> str:
    current_time = datetime.datetime.utcnow()
    time_ago = current_time - dt

    # Convert the time difference to a readable string
    if time_ago < datetime.timedelta(minutes=1):
        time_ago_str = "just now"
    elif time_ago < datetime.timedelta(hours=1):
        minutes = int(time_ago.total_seconds() / 60)
        time_ago_str = f"{minutes} minutes ago"
    elif time_ago < datetime.timedelta(days=1):
        hours = int(time_ago.total_seconds() / 3600)
        time_ago_str = f"{hours} hours ago"
    else:
        days = time_ago.days
        time_ago_str = f"{days} days ago"

    return time_ago_str


def timedelta(time_span: str) -> datetime.timedelta:
    hour_pattern = re.compile(r"(\d+)h")
    day_pattern = re.compile(r"(\d+)d")
    week_pattern = re.compile(r"(\d+)w")
    if m := hour_pattern.match(time_span):
        return datetime.timedelta(hours=int(m.group(1)))
    elif m := day_pattern.match(time_span):
        return datetime.timedelta(days=int(m.group(1)))
    elif m := week_pattern.match(time_span):
        return datetime.timedelta(weeks=int(m.group(1)))
    raise ValueError("Invalid time frame")


class LinkStyle:
    def __init__(self, scheme: str | None = None):
        # ivory://acct/openURL?url=
        # {config.ConfigHandler.MASTO_BASE}/deck/@{toot.author}/{toot.toot_id}
        if scheme:
            self.scheme = st.radio("Link scheme:", ["Desktop", "Ivory", "Original"], index=1, horizontal=True)
        else:
            self.scheme = scheme

    def toot_url(self, toot: core.Toot) -> str:
        if self.scheme == "Desktop":
            return f"{config.ConfigHandler.MASTO_BASE}/@{toot.author}/{toot.toot_id}"
        elif self.scheme == "Ivory":
            encoded_url = urllib.parse.quote(toot.url)
            return f"ivory://acct/openURL?url={encoded_url}"
        elif self.scheme == "Original":
            return toot.url
        raise ValueError("Invalid scheme")

    def profile_url(self, toot: core.Toot) -> str:
        if self.scheme == "Desktop":
            return f"{config.ConfigHandler.MASTO_BASE}/@{toot.author}"
        elif self.scheme == "Ivory":
            # return f"ivory://@{toot.author}/profile"
            return f"ivory://acct/openURL?url={toot.profile_url}"
        elif self.scheme == "Original":
            return toot.profile_url
        raise ValueError("Invalid scheme")



class TootCluster(pydantic.BaseModel):
    id: int
    name: str
    toots: list[core.Toot]


class TootClusters(pydantic.BaseModel):
    clusters: list[TootCluster]

    @property
    def num_toots(self) -> int:
        return sum(len(c.toots) for c in self.clusters)

    @property
    def max_date(self) -> datetime.datetime:
        seq = [t.created_at for c in self.clusters for t in c.toots]
        return max(seq) if len(seq) > 0 else datetime.datetime.utcnow()

    @property
    def min_date(self) -> datetime.datetime:
        seq = [t.created_at for c in self.clusters for t in c.toots]
        return min(seq) if len(seq) > 0 else datetime.datetime.utcnow()


def display_toot(toot: core.Toot, link_style: LinkStyle):
    with st.container(border=True):
        reply = "↩" if toot.is_reply else ""
        st.markdown(f"""
{reply}<a href="{link_style.profile_url(toot)}"><img src="{toot.avatar_url}" style="width: 40px; height: 40px" />{toot.display_name} @{toot.author} ({time_ago(toot.created_at)})</a>
{toot.content}
""", unsafe_allow_html=True)

        attachments = [f'<a href="{a.url}"><img src="{a.preview_url}" style="max-width: 100%" /></a>' for a in toot.media_attachments]
        st.markdown(" ".join(attachments), unsafe_allow_html=True)

        cols = st.columns(4)
        with cols[0]:
            st.markdown(f"""<a href="{link_style.toot_url(toot)}" target="_blank">🔗</a>""", unsafe_allow_html=True)
        with cols[1]:
            if st.button("⭐️", key=f"star-{toot.id}"):
                toot.do_star()
        with cols[2]:
            if st.button("️🔁", key=f"boost-{toot.id}"):
                toot.do_boost()
        with cols[3]:
            if st.button("🪲", key=f"delete-{toot.id}"):
                import json
                print(json.dumps(toot.orig_dict, indent=2))


def all_toot_summary(toots: list[core.Toot]):
    latest_date = max(t.created_at for t in toots)
    earliest_date = min(t.created_at for t in toots)
    now = datetime.datetime.utcnow()
    msg = f"{len(toots)} toots from {time_ago(earliest_date)} to {time_ago(latest_date)}"
    if latest_date > now:
        st.warning(msg)
    else:
        st.info(msg)


"""
A FastAPI HTML server.

The streamlit version had issues around state management and was genrally slow
and inflexible. This gives us a lot more control.
"""
import datetime
import importlib
import json
from typing import Annotated

import llm
import requests
from fastapi import FastAPI, Form, HTTPException, Request, responses, staticfiles, templating

from fossil_mastodon import algorithm, config, core, plugins, ui


app = FastAPI()


app.mount("/static", staticfiles.StaticFiles(directory=config.ASSETS.assets_path), name="static")
templates = templating.Jinja2Templates(directory=config.ASSETS.templates_path)
print("using template directory", config.ASSETS.templates_path)
templates.env.filters["rel_date"] = ui.time_ago


@app.middleware("http")
async def session_middleware(request: Request, call_next):
    session_id = request.cookies.get("fossil_session_id")
    session = core.Session.get_by_id(session_id) if session_id else None
    if session is None:
        session = core.Session.create()
        session.save()
        request.state.session = session
        response = await call_next(request)
        response.set_cookie("fossil_session_id", session.id)  
        return response
    else:
        request.state.session = session
        return await call_next(request)


@app.get("/")
async def root(request: Request):
    session: core.Session = request.state.session
    ctx = plugins.RenderContext(
        templates=templates,
        request=request,
        link_style=ui.LinkStyle("Desktop"),
        session=session,
    )

    algo_list = plugins.get_algorithms()
    if len(algo_list) == 0:
        print(f"No algorithms found (num plugins: {len(plugins.get_plugins())})")
        for plugin in plugins.get_plugins():
            print(f"Plugin ({plugin.name})", plugin)
        return templates.TemplateResponse("no_algorithm.html", {
            "request": request,
        })

    algo = session.get_algorithm_type() or algo_list[0]
    return templates.TemplateResponse("index.html", {
        "request": request,
        "model_params": algo.render_model_params(ctx).body.decode("utf-8"),
        "ui_settings": session.get_ui_settings(),
        "selected_algorithm": algo,
        "algorithms": [
            {"name": algo.plugin.name, "display_name": algo.plugin.display_name}
            for algo in plugins.get_algorithms()
        ],
    })


@app.get("/toots")
async def toots():
    return staticfiles.FileResponse("public/toots.html")


@app.post("/toots/download")
async def toots_download(request: Request):
    core.create_database()
    session: core.Session = request.state.session
    core.download_timeline(datetime.datetime.utcnow() - datetime.timedelta(days=1), session.id)
    algorithm_spec = json.loads(session.algorithm_spec) if session.algorithm_spec else {}
    body_params: dict[str, str] = dict((await request.form()))
    session.set_ui_settings(body_params)
    print("algorithm_spec", algorithm_spec)
    if "module" in algorithm_spec and "class_name" in algorithm_spec:
        mod = importlib.import_module(algorithm_spec["module"])
        model: algorithm.BaseAlgorithm = getattr(mod, algorithm_spec["class_name"]).train(
            algorithm.TrainContext(
                end_time=datetime.datetime.utcnow(),
                timedelta=datetime.timedelta(days=1),
                session_id=session.id
            ),
            algorithm_spec["kwargs"],
        )
        timespan = ui.timedelta(body_params["time_span"])
        timeline = core.Toot.get_toots_since(datetime.datetime.utcnow() - timespan)
        renderable = model.render(timeline, plugins.RenderContext(
            templates=templates,
            request=request,
            link_style=ui.LinkStyle(body_params["link_style"] if "link_style" in body_params else "Desktop"),
            session=session,
        ))
        return renderable.render()
    else:
        return responses.HTMLResponse("<div>No Toots 😥</div>")


@app.post("/toots/train")
async def toots_train(
    link_style: Annotated[str, Form()],
    time_span: Annotated[str, Form()],
    request: Request,
):
    context = algorithm.TrainContext(
        end_time=datetime.datetime.utcnow(),
        timedelta=ui.timedelta(time_span),
        session_id=request.state.session.id
    )

    algo_kwargs = {k: v for k, v in dict((await request.form())).items() 
                   if k not in {"link_style", "time_span"}}
    print("Algorithm kwargs:", algo_kwargs)

    # train
    session: core.Session = request.state.session
    algo = session.get_algorithm_type() or plugins.get_algorithms()[0]
    model = algo.train(context, algo_kwargs)
    session.algorithm = model.serialize()
    session.algorithm_spec = json.dumps({
        "module": model.__class__.__module__,
        "class_name": model.__class__.__qualname__,
        "kwargs": algo_kwargs,
    })
    session.save()

    # render
    timeline = core.Toot.get_toots_since(datetime.datetime.utcnow() - ui.timedelta(time_span))
    renderable = model.render(timeline, plugins.RenderContext(
        templates=templates,
        request=request,
        link_style=ui.LinkStyle(link_style),
        session=session,
    ))
    try:
        return renderable.render()
    except plugins.BadPluginFunction as ex:
        return templates.TemplateResponse("bad_plugin.html", { "request": request, "ex": ex })


@app.get("/algorithm/{name}/form")
async def algorithm_form(name: str, request: Request):
    session: core.Session = request.state.session
    algo_type = session.get_algorithm_type() or plugins.get_algorithms()[0]
    ctx = plugins.RenderContext(
        templates=templates,
        request=request,
        link_style=ui.LinkStyle(session.get_ui_settings().get("link_style", "Desktop")),
        session=session,
    )
    return algo_type.render_model_params(ctx)


@app.get("/settings")
async def get_settings(request: Request):
    session: core.Session = request.state.session
    keys = {"openai": "", **llm.load_keys()}
    return templates.TemplateResponse("settings.html", {
        "request": request,
        "settings": session.settings,
        "embedding_models": config.get_installed_embedding_models(),
        "embedding_model": session.settings.embedding_model,
        "summarize_models": config.get_installed_llms(),
        "summarize_model": session.settings.summarize_model,
        "keys": keys,
    })

@app.post("/settings")
async def post_settings(settings: core.Settings, request: Request):
    session: core.Session = request.state.session
    session.settings = settings
    session.save()
    return responses.HTMLResponse("<div>👍</div>")

@app.post("/keys")
async def post_keys(request: Request):
    body_params: dict[str, str] = dict((await request.form()))
    key_path = llm.user_dir() / "keys.json"
    key_path.write_text(json.dumps(body_params))
    return responses.HTMLResponse("<div>👍</div>")


@app.post("/toots/{id}/debug")
async def toots_debug(id: int):
    toot = core.Toot.get_by_id(id)
    if toot is not None:
        import json
        print(json.dumps(toot.orig_dict, indent=2))
    return responses.HTMLResponse("<div>💯</div>")

@app.post("/toots/{id}/boost")
async def toots_boost(id: int):
    toot = core.Toot.get_by_id(id)
    if toot is not None:
        url = f'{config.ConfigHandler.MASTO_BASE}/api/v1/statuses/{toot.toot_id}/reblog'
        data = {
            'visibility': 'public'
        }
        response = requests.post(url, json=data, headers=config.headers())
        try:
            response.raise_for_status()
            return responses.HTMLResponse("<div>🚀</div>")
        except:
            print("ERROR:", response.json())
            raise
    raise HTTPException(status_code=404, detail="Toot not found")

@app.post("/toots/{id}/favorite")
async def toots_favorite(id: int):
    toot = core.Toot.get_by_id(id)
    if toot is not None:
        url = f'{config.ConfigHandler.MASTO_BASE}/api/v1/statuses/{toot.toot_id}/favourite'
        response = requests.post(url, headers=config.headers())
        try:
            response.raise_for_status()
            return responses.HTMLResponse("<div>💫</div>")
        except:
            print("ERROR:", response.json())
            raise
    raise HTTPException(status_code=404, detail="Toot not found")
    

# this should always be the last line of this file
plugins.init_plugins(app)


import abc
import functools
import inspect
import re
import traceback
from typing import Callable, Type

from fastapi import FastAPI, Request, responses, templating
import pkg_resources
import pydantic

from fossil_mastodon import algorithm, ui, core


def title_case_to_spaced(string):
    # The regex pattern looks for any lowercase letter followed by an uppercase letter
    # and inserts a space between them
    return re.sub(r'(?<=[a-z])(?=[A-Z])', ' ', string)


TootDisplayFn = Callable[[core.Toot, "RenderContext"], responses.Response]
class TootDisplayPlugin(pydantic.BaseModel):
    fn: TootDisplayFn
    fn_name: str

    def render_str(self, toot: core.Toot, context: "RenderContext") -> str:
        obj = self.fn(toot, context)
        content = obj.body.decode("utf-8")
        return content


class RenderContext(pydantic.BaseModel):
    """
    A context object for rendering a template.
    """
    class Config:
        arbitrary_types_allowed = True
    templates: templating.Jinja2Templates
    request: Request
    link_style: ui.LinkStyle
    session: core.Session

    def template_args(self) -> dict:
        return {
            "request": self.request,
            "link_style": self.link_style,
            "ctx": self,
        }

    def render_toot_display_plugins(self, toot: core.Toot) -> str:
        return "".join(
            plugin.render_str(toot, self)
            for plugin in get_toot_display_plugins()
        )


_app: FastAPI | None = None


class Plugin(pydantic.BaseModel):
    """
    Plugin registration API
    
    Example:

        plugin = Plugin(name="My Plugin", description="Add button to toot that triggers an API POST operation")

        @plugin.api_operation.post("/my_plugin")
        def my_plugin(request: Request):
            return responses.HTMLResponse("<div>💯</div>")

        @plugin.toot_display_button
        def my_toot_display(toot: core.Toot, context: RenderContext):
            return responses.HTMLResponse("<div>💯</div>")

    """
    name: str
    display_name: str | None = None
    description: str | None = None
    author: str | None = None
    author_url: str | None = None
    enabled_by_default: bool = True
    _toot_display_buttons: list[TootDisplayPlugin] = pydantic.PrivateAttr(default_factory=list)
    _algorithms: list[Type[algorithm.BaseAlgorithm]] = pydantic.PrivateAttr(default_factory=list)

    @pydantic.validator("display_name", always=True)
    def _set_display_name(cls, v, values):
        return v or values["name"]

    @property
    def api_operation(self) -> FastAPI:
        assert _app is not None
        return _app

    def toot_display_button(self, impl: TootDisplayFn) -> TootDisplayFn:
        """
        Decorator for adding a button to the toot display UI. This function should return a
        fastapi.responses.Response object. The result will be extracted and inserted into the
        toot display UI.
        """
        name = impl.__name__

        @functools.wraps(impl)
        def wrapper(toot: core.Toot, context: RenderContext):
            try:
                return impl(toot, context)
            except TypeError as e:
                raise BadPluginFunction(self, impl, "example_function(toot: fossil_mastodon.core.Toot, context: fossil_mastodon.plugins.RenderContext)") from e
            except Exception as e:
                import inspect
                print(inspect.signature(impl))
                raise RuntimeError(f"Error in toot display plugin '{self.name}', function '{name}'") from e

        self._toot_display_buttons.append(TootDisplayPlugin(fn=wrapper, fn_name=name))
        return wrapper

    def algorithm(self, algo: Type[algorithm.BaseAlgorithm]) -> Type[algorithm.BaseAlgorithm]:
        """
        Decorator for adding an algorithm class.
        """
        if not issubclass(algo, algorithm.BaseAlgorithm):
            raise ValueError(f"Algorithm {algo} is not a subclass of algorithm.BaseAlgorithm")
        self._algorithms.append(algo)
        algo.plugin = self
        return algo


def init_plugins(app: FastAPI):
    global _app
    _app = app
    get_plugins()


@functools.lru_cache
def get_plugins() -> list[Plugin]:
    if _app is None:
        raise RuntimeError("Plugins not initialized")

    plugins = []
    for entry_point in pkg_resources.iter_entry_points("fossil_mastodon.plugins"):
        print("Loading plugin", entry_point.name)
        try:
            plugin = entry_point.load()
            if isinstance(plugin, Plugin):
                plugins.append(plugin)
            else:
                print(f"Error loading toot display plugin '{entry_point.name}': not a subclass of Plugin")
        except:
            print(f"Error loading toot display plugin {entry_point.name}")
            traceback.print_exc()
    return plugins


def get_toot_display_plugins() -> list[TootDisplayPlugin]:
    return [
        b 
        for p in get_plugins() 
        for b in p._toot_display_buttons
    ]


def get_algorithms() -> list[Type[algorithm.BaseAlgorithm]]:
    return [
        algo
        for p in get_plugins()
        for algo in p._algorithms
    ]


class BadPluginFunction(Exception):
    def __init__(self, plugin: Plugin, function: callable, expected_signature: str):
        super().__init__(f"Bad function call: {plugin.name}.{function.__name__} should have signature {expected_signature}")
        self.plugin = plugin
        self.function = function
        self.signature = inspect.signature(function)
        self.expected_signature = expected_signature
        self.function_name = function.__name__



import llm
import numpy as np
import pydantic
import tiktoken
from fastapi import Response, responses
from sklearn.cluster import KMeans
from tqdm import trange

from fossil_mastodon import config, core, plugins, ui
from fossil_mastodon import algorithm


plugin = plugins.Plugin(
    name="Topic Cluster",
    description="Cluster toots by topic",
)


class ClusterRenderer(algorithm.Renderable, pydantic.BaseModel):
    clusters: list[ui.TootCluster]
    context: plugins.RenderContext

    def render(self, **response_args) -> Response:
        toot_clusters = ui.TootClusters(clusters=self.clusters)
        return self.context.templates.TemplateResponse("toot_clusters.html", {
            "clusters": toot_clusters,
            **self.context.template_args(),
        },
        **response_args)


@plugin.algorithm
class TopicCluster(algorithm.BaseAlgorithm):
    def __init__(self, kmeans: KMeans, labels: dict[int, str]):
        self.kmeans = kmeans
        self.labels = labels

    def render(self, toots: list[core.Toot], context: plugins.RenderContext) -> ClusterRenderer:
        before = len(toots)
        toots = [toot for toot in toots if toot.embedding is not None]
        print("Removed", before - len(toots), "toots with no embedding (probably image-only).", f"{len(toots)} toots remaining.")
        cluster_indices = self.kmeans.predict(np.array([toot.embedding for toot in toots]))
        for toot, cluster_index in zip(toots, cluster_indices):
            toot.cluster = self.labels[cluster_index]

        toot_clusters = ui.TootClusters(
            clusters=[
                ui.TootCluster(
                    id=i_cluster,
                    name=cluster_label,
                    toots=[toot for toot, cluster_index in zip(toots, cluster_indices) if cluster_index == i_cluster],
                )
                for i_cluster, cluster_label in self.labels.items()
            ]
        )
        return ClusterRenderer(clusters=toot_clusters.clusters, context=context)

    @classmethod
    def train(cls, context: algorithm.TrainContext, args: dict[str, str]) -> "TopicCluster":
        toots = [toot for toot in context.get_toots() if toot.embedding is not None]
        print("lengths:", {len(toot.embedding) for toot in toots})

        n_clusters = int(args["num_clusters"])
        if len(toots) < n_clusters:
            return cls(kmeans=NoopKMeans(n_clusters=1), labels={0: "All toots"})

        embeddings = np.array([toot.embedding for toot in toots])
        kmeans = KMeans(n_clusters=n_clusters)
        cluster_labels = kmeans.fit_predict(embeddings)

        labels: dict[int, str] = {}
        model = llm.get_model(config.ConfigHandler.SUMMARIZE_MODEL(context.session_id).name)
        for i_clusters in trange(n_clusters):
            clustered_toots = [toot for toot, cluster_label in zip(toots, cluster_labels) if cluster_label == i_clusters]
            combined_text = "\n\n".join([toot.content for toot in clustered_toots])

            # Use the summarizing model to summarize the combined text
            prompt = f"Create a single label that describes all of these related tweets, make it succinct but descriptive. The label should describe all {len(clustered_toots)} of these\n\n{combined_text}"
            summary = model.prompt(reduce_size(context.session_id, prompt)).text().strip()
            labels[i_clusters] = summary

        return cls(kmeans=kmeans, labels=labels)

    @staticmethod
    def render_model_params(context: plugins.RenderContext) -> Response:
        default = context.session.get_ui_settings().get("num_clusters", "15")
        return responses.HTMLResponse(f"""
            <div class="slider">
                <input type="range" name="num_clusters" id="num_clusters" min="0" max="20" value="{default}" onchange="document.getElementById('num_clusters_value').innerHTML = this.value">
                <span><span class="slider-value" id="num_clusters_value">{default}</span> clusters</span>
            </div>
        """)

def get_encoding(session_id: str):
    try:
        return tiktoken.encoding_for_model(config.ConfigHandler.SUMMARIZE_MODEL(session_id).name)
    except KeyError:
        encoding_name = tiktoken.list_encoding_names()[-1]
        return tiktoken.get_encoding(encoding_name)

def reduce_size(session_id: str, text: str, model_limit: int = -1, est_output_size: int = 500) -> str:
    if model_limit < 0:
        model_limit = config.ConfigHandler.SUMMARIZE_MODEL(session_id).context_length
    tokens = get_encoding(session_id).encode(text)
    return get_encoding(session_id).decode(tokens[:model_limit - est_output_size])


class NoopKMeans(KMeans):
    def predict(self, X, y=None, sample_weight=None):
        return np.zeros(len(X), dtype=int)



from fastapi import responses

from fossil_mastodon import plugins, core


plugin = plugins.Plugin(
    name="Toot Debug Button",
    description="Adds a button to toots that prints the toot's JSON to the server's console.",
)


@plugin.api_operation.post("/plugins/toot_debug/{id}")
async def toots_debug(id: int):
    toot = core.Toot.get_by_id(id)
    if toot is not None:
        import json
        print(json.dumps(toot.orig_dict, indent=2))
    return responses.HTMLResponse("<div>💯</div>")


@plugin.toot_display_button
def get_response(toot: core.Toot, context: plugins.RenderContext) -> responses.Response:
    return responses.HTMLResponse(f"""
        <button hx-post="/plugins/toot_debug/{ toot.id }">🪲</button>
    """)